import ipywidgets as widgets
from IPython.display import display, HTML

from ..logic.vis import histogram


class EDAHubWidgetColumnSummary:
    def __init__(self):
        self.output = widgets.Accordion()

    def update(self, edahub):
        table_names = edahub.get_table_names()
        contents = []
        titles = []
        for table_name in table_names:
            stats = widgets.Output(layout=widgets.Layout(width='100%'))
            histogram = widgets.Output(layout=widgets.Layout(width='100%'))
            select_chart, filter_text = self._get_histogram_selection(table_name, edahub, histogram)
            if select_chart.options:
                select_chart.value = select_chart.options[0]
            with stats:
                display(HTML(self._get_table_html(table_name, edahub.stats_tables[table_name])))
            contents.append(
                widgets.VBox(
                    [
                        stats, 
                        widgets.HBox([select_chart, filter_text]),
                        histogram
                    ]
                )
            )
            titles.append(table_name)
        self.output.children = contents
        for i, title in enumerate(titles):
            self.output.set_title(i, title)

    def _get_table_html(self, table_name, table_df):
        return table_df.to_html()

    def _get_histogram_selection(self, table_name, edahub, histogram):
        available_columns = list(edahub.histograms[table_name].keys()) if table_name in edahub.histograms \
            else []
        select_chart = widgets.Select(
            options=available_columns,
            description='Select Chart:',
            disabled=False
        )
        filter_text = widgets.Text(
            placeholder='',
            description='Type to filter:',
            disabled=False
        )

        def update_chart(change):
            column_name = change['new']
            chart = edahub.histograms[table_name].get(column_name)
            if chart:
                with histogram:
                    histogram.clear_output()
                    display(chart)

        def filter_options(text):
            filtered_options = [option for option in available_columns if text.lower() in option.lower()]
            select_chart.options = filtered_options

        def handle_text_change(change):
            filter_options(change['new'])

        select_chart.observe(update_chart, names='value')
        filter_text.observe(handle_text_change, names='value')
        return select_chart, filter_text



from datetime import datetime
import threading

import pandas

from ..widgets import EDAHubWidget
from . import stats_calculator
from .vis import histogram


STATUS_COLUMNS = ["name", "#rows", "#cols", "status", "created"]


class EDAHub:
    def __init__(self, name=None):
        self.name = name or _get_name()
        self.tables = {}
        self.status_table = pandas.DataFrame(columns=STATUS_COLUMNS)
        self.stats_tables = {}
        self.histograms = {}
        self.eda_thread = None
        self.widget = EDAHubWidget(self)

    def add_table(self, name, df, overwrite=False):
        if not overwrite and name in self.tables:
            raise Exception(f"name={name} has been already registered. Use a different name or the option `overwrite=True`")
        now = datetime.now().replace(microsecond=0)
        new_row = pandas.DataFrame(
            {k: [v] for k, v in zip(
                STATUS_COLUMNS, [name, len(df), len(df.columns), "computing stats", now]
            )}
        )
        if len(self.status_table):
            self.status_table = pandas.concat([self.status_table, new_row], ignore_index=True)
        else:
            self.status_table = new_row
        self.tables[name] = df
        self.stats_tables[name] = self._get_schema(df)
        self.widget.update(None)
        self._compute_stats(name, df)
        self._add_histograms(name, df)

    def get_table_names(self):
        #TODO: sort table names based on created
        return list(self.tables.keys())

    def _get_schema(self, df):
        return pandas.DataFrame({
            "column_name": df.columns,
            "dtype": df.dtypes
        })

    def _compute_stats(self, name, df):
        def target():
            self.stats_tables[name] = stats_calculator.calc_column_stat(df)
            self._update_status(name, "generating histograms")
            self.widget.update(None)
        
        self.eda_thread = threading.Thread(target=target)
        self.eda_thread.start()

    def _add_histograms(self, name, df):
        def target():
            self.histograms[name] = {
                c: histogram.plot_histogram(self.tables[name][c])
                for c in self.tables[name].columns
            }
            self._update_status(name, "done")
            self.widget.update(None)
        
        self.eda_thread = threading.Thread(target=target)
        self.eda_thread.start()

    def _update_status(self, name, status):
        self.status_table.loc[self.status_table["name"]==name, "status"] = status


def _get_name():
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S')



import pandas
from sidecar import Sidecar
import ipywidgets as widgets
import time
from IPython.display import display, HTML

from .column_summary import EDAHubWidgetColumnSummary


class EDAHubWidget:
    def __init__(self, edahub):
        self.edahub = edahub
        tab_contents = ["Summary", "Schema & Stats", "Warnings"]
        self.sidecar = Sidecar(title=f'EDAHub({self.edahub.name})', anchor='right')
        self.update_button = widgets.Button(description="Update")
        self.update_button.on_click(self.update)
        self.summary = EDAHubWidgetSummary()
        self.column_summary = EDAHubWidgetColumnSummary() 
        self.warnings = EDAHubWidgetWarnings()
        with self.sidecar:
            tab = widgets.Tab()
            tab.children=[
                self.summary.output,
                self.column_summary.output,
                self.warnings.output
            ]
            tab.titles = tab_contents
            display(widgets.VBox([self.update_button, tab]))

    def update(self, _):
        self.summary.update(self.edahub)
        self.column_summary.update(self.edahub)


class EDAHubWidgetSummary:
    def __init__(self):
        self.status_text = widgets.Output(layout=widgets.Layout(width='100%'))
        self.status_table = widgets.Output(layout=widgets.Layout(width='100%'))
        self.output = widgets.VBox(
            [self.status_text, self.status_table]
        )

    def update(self, edahub):
        with self.status_text:
            text = f"{len(edahub.tables)} tables"
            self.status_text.clear_output()
            display(HTML(f"<p>{text}</p>"))
            
        with self.status_table:
            self.status_table.clear_output()
            display(HTML(edahub.status_table.to_html()))




class EDAHubWidgetWarnings:
    def __init__(self):
        self.output = widgets.Accordion(children=[])
